// Copyright (c) 2019, Christoph Peters
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the Karlsruhe Institute of Technology nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include "TristimulusIntegrator.h"
#include "Grid3D.h"
#include "NearestNeighborIndex.h"
#include "colorSpaceConversion.h"
#include "XYZWarp.h"
#include <iostream>
#include <memory>
#include <utility>
#include <fstream>
#include <thread>
#include <string>

/*! Generates a grid of the given resolution covering the entire relevant range
	of Fourier coefficients for reflectance spectra and fills it with the
	corresponding XYZ triples. mirrorSignal forwards to FourierSpectrum.
	\param useFourierSRGB Pass false to obtain a grid sampling the whole range
			of valid Fourier coefficients. Pass true to obtain a Fourier sRGB grid
			with minimal values 0.0f and maximal values 1.0f.
	\sa convertMirroredWarpedFourierToFourierXYZ()
	\sa convertXYZToSRGB() */
template<bool mirrorSignal, bool useBiasing>
std::unique_ptr<Grid3D> getFourierToXYZGrid(std::array<std::size_t, 3> resolution, FourierColorSpace fourierToXYZDomain) {
	// Define integrator and wavelength mapping
	TristimulusIntegrator integrator = getXYZIntegrator(5);
	auto wavelengthToPhase = [](float wavelength, float& rOutPhase)->bool {
		if (mirrorSignal) {
			rOutPhase = applyXYZWarpEven(wavelength);
		}
		else {
			rOutPhase = applyXYZWarpOdd(wavelength);
		}
		return (wavelength >= 360.0f && wavelength <= 830.0f);
	};
	// Initialize the grid
	std::unique_ptr<Grid3D> pResult;
	if (fourierToXYZDomain == FourierColorSpace::fourierSRGB) {
		std::array<float, 3> expandedGridMin = { 0.0f, 0.0f, 0.0f };
		std::array<float, 3> expandedGridMax = { 1.0f, 1.0f, 1.0f };
		Grid3D::expandGrid(resolution, expandedGridMin, expandedGridMax);
		pResult.reset(new Grid3D(resolution, expandedGridMin, expandedGridMax));
	}
	else if (fourierToXYZDomain == FourierColorSpace::fourierXYZ) {
		// These boundary values have been computed in Python by integrating
		// each basis function over the regions where it is negative / positive
		pResult.reset(new Grid3D(resolution, {0.0f, -2.34760958f, -5.62733921f}, {107.10574724f, 109.20459766f, 114.5296003f}));
	}
	else {
		pResult.reset(new Grid3D(resolution, { 0.0f, -1.0f / 3.14159265358979f, -1.0f / 3.14159265358979f }, { 1.0f, 1.0f / 3.14159265358979f, 1.0f / 3.14159265358979f }));
	}
	// Fill it with data
	pResult->fill(std::thread::hardware_concurrency(),
		[&integrator, &wavelengthToPhase, fourierToXYZDomain](const std::array<float, 3>& rFourierCoefficients)->std::array<float, 3>
	{
		std::array<float, 3> plainFourier = convertAnyFourierToMirroredWarpedFourier(rFourierCoefficients, fourierToXYZDomain);
		return integrator.integrate(FourierSpectrum<mirrorSignal, useBiasing, decltype(wavelengthToPhase)>(wavelengthToPhase, plainFourier));
	}
	);
	return pResult;
}


/*! Turns a grid such as the one produced by getFourierToXYZGrid() into an
	inverse grid mapping each XYZ triple to three Fourier coefficients. This is
	done by means of nearest neighbor lookups of the data in the given grid
	using the Euclidean norm in the CIELAB color space.
	\param fourierToXYZDomain The domain of rFourierToXYZGrid.
	\param xyzToFourierDomain The domain of the returned grid.
	\param xyzToFourierRange The range of the returned grid, i.e. the color
			space of output values.*/
std::unique_ptr<Grid3D> getXYZToFourierGrid(std::array<std::size_t, 3> resolution, const Grid3D& rFourierToXYZGrid, FourierColorSpace fourierToXYZDomain, TristimulusColorSpace xyzToFourierDomain, FourierColorSpace xyzToFourierRange) {
	std::array<float, 3> cuboidMin = { 0.0f, 0.0f, 0.0f };
	std::array<float, 3> cuboidMax = { 1.0f, 1.0f, 1.0f };
	if (xyzToFourierDomain == TristimulusColorSpace::xyz) {
		// Determine the maximal valid XYZ values by integrating the brightest 
		// reflectance spectrum
		TristimulusIntegrator integrator = getXYZIntegrator(1);
		cuboidMax = integrator.integrateConstantOne();
	}
	else {
		// For sRGB, we want to construct the lookup table such that quantized 
		// sRGB provides indices into the table. Commonly the smallest 
		// quantized value means 0.0 and the largest means 1.0, so we have to 
		// adapt the grid spacing.
		Grid3D::expandGrid(resolution, cuboidMin, cuboidMax);
	}
	// Define the grid
	std::unique_ptr<Grid3D> pResult;
	pResult.reset(new Grid3D(resolution, cuboidMin, cuboidMax));
	// Turn the XYZ in the source grid into CIELAB and prepare the lookup 
	// structure
	std::cout << "Converting XYZ to CIELAB..." << std::endl;
	const auto& rPointListXYZ = rFourierToXYZGrid.getData();
	std::vector<std::array<float, 3>> pointListLAB(rFourierToXYZGrid.getCellCount());
	std::transform(rPointListXYZ.begin(), rPointListXYZ.end(), pointListLAB.begin(), 
		[](const std::array<float, 3>& rXYZ) -> std::array<float, 3> {
			return convertXYZToCIELAB(rXYZ, false);
		}
	);
	std::cout << "Building acceleration structure for nearest neighbor lookups..." << std::endl;
	NearestNeighborIndex nearestNeighborIndex(pointListLAB, { -1.0f, -1.0f, -1.0f });
	// Fill it with data
	std::cout << "Filling the output grid..." << std::endl;
	pResult->fill(std::thread::hardware_concurrency(),
		[&nearestNeighborIndex, &rFourierToXYZGrid, fourierToXYZDomain, xyzToFourierDomain, xyzToFourierRange](const std::array<float, 3>& rTristimulusCoefficients)->std::array<float, 3>
	{
		std::array<float, 3> xyz = rTristimulusCoefficients;
		if (xyzToFourierDomain == TristimulusColorSpace::sRGB) {
			xyz = convertSRGBToXYZ(rTristimulusCoefficients, false);
		}
		std::size_t index = nearestNeighborIndex.query(convertXYZToCIELAB(xyz, false));
		std::array<float, 3> fourier = rFourierToXYZGrid.getDataCoordinate(rFourierToXYZGrid.unravelIndex(index));
		return convertAnyFourierToAnyFourier(fourier, fourierToXYZDomain, xyzToFourierRange);
	}
	);
	return pResult;
}


/*! Determines an appropriate normalization factor for the definition of the
	Fourier XYZ color space. It is defined such that Y and Fourier Y match
	exactly for constant spectra. The result is hard-coded in
	convertMirroredWarpedFourierToFourierXYZ(), so this function should return
	1.0f.*/
float computeFourierXYZNormalizationFactor() {
	TristimulusIntegrator integrator = getXYZIntegrator();
	// XYZ coefficients of a spectrum that is constant at one
	std::array<float, 3> xyz = integrator.integrateConstantOne();
	// Fourier coefficients of a spectrum that is constant at one
	std::array<float, 3> fourier = { 1.0f, 0.0f, 0.0f };
	std::array<float, 3> fourierXYZ = convertMirroredWarpedFourierToFourierXYZ(fourier);
	return xyz[1] / fourierXYZ[1];
}


int main(int nArgument, char* ppArgument[]) {
	// The resolution of the grid mapping from a Fourier color space to XYZ
	std::size_t fourierToXYZResolution = 1024;
	// The resolution of the grid mapping from a tristimulus color space to a
	// Fourier color space
	std::size_t xyzToFourierResolution = 256;
	// First you have to generate a grid from Fourier to XYZ, then the reverse
	// lookup table can be constructed from that
	bool makeFourierToXYZGrid = true;
	// The Fourier color space used to construct the Fourier to XYZ grid. Note
	// that Fourier sRGB gives an incomplete grid but helps to get a more dense
	// sampling for dark colors.
	FourierColorSpace fourierToXYZDomain = FourierColorSpace::plain;
	// The tristimulus color space used to construct the grid mapping a
	// tristimulus color space to Fourier coefficients
	TristimulusColorSpace xyzToFourierDomain = TristimulusColorSpace::xyz;
	// The Fourier color space used to store output values of the grid mapping
	// a tristimulus color space to Fourier coefficients
	FourierColorSpace xyzToFourierRange = FourierColorSpace::plain;
	// Construct file names
	std::string fourierToXYZFileName = getColorSpaceName(fourierToXYZDomain) + "To" + getColorSpaceName(TristimulusColorSpace::xyz) + std::to_string(fourierToXYZResolution) + "Even.dat";
	std::string xyzToFourierFileName = getColorSpaceName(xyzToFourierDomain) + "To" + getColorSpaceName(xyzToFourierRange) + std::to_string(xyzToFourierResolution) + "Even.dat";
	if (makeFourierToXYZGrid) {
		auto pXYZToFourier = getFourierToXYZGrid<true, false>({ fourierToXYZResolution, fourierToXYZResolution, fourierToXYZResolution }, fourierToXYZDomain);
		std::ofstream outputStream(fourierToXYZFileName, std::ios_base::binary);
		pXYZToFourier->store(outputStream);
	}
	else {
		std::ifstream inputStream(fourierToXYZFileName, std::ios_base::binary);
		if (!inputStream.is_open()) {
			std::cout << "The input file " << fourierToXYZFileName << " could not be opened. Please generate it first." << std::endl;
			return 0;
		}
		std::cout << "Loading " << fourierToXYZFileName << "..." << std::endl;
		Grid3D fourierToXYZ(inputStream);
		auto pFourierToXYZ = getXYZToFourierGrid({ xyzToFourierResolution, xyzToFourierResolution, xyzToFourierResolution }, fourierToXYZ, fourierToXYZDomain, xyzToFourierDomain, xyzToFourierRange);
		std::cout << "Storing the grid to " << xyzToFourierFileName << "..." << std::endl;
		std::ofstream outputStream(xyzToFourierFileName, std::ios_base::binary);
		pFourierToXYZ->store(outputStream);
	}
	std::cout << "Done." << std::endl;
	return 0;
}
